import math
from sample.maze.maze_domain import MazeDomain
from kyoka.value_function.base_keras_action_value_function import BaseKerasActionValueFunction
from keras.models import Sequential
from keras.layers.core import Dense, Activation
import numpy as np

class MazeKerasValueFunction(BaseKerasActionValueFunction):

  def __init__(self, domain):
    BaseKerasActionValueFunction.__init__(self)
    self.domain = domain
    self.maze_shape = domain.get_maze_shape()

  def generate_model(self):
    input_dim = self.maze_shape[0] * self.maze_shape[1]
    model = Sequential()
    model.add(Dense(100, input_shape=(input_dim,)))
    model.add(Activation("relu"))
    model.add(Dense(1))
    model.compile(loss="mse",  optimizer="adam")
    return model

  def transform_state_action_into_input(self, state, action):
    flatten = lambda l: [item for sublist in l for item in sublist]
    next_state = self.domain.transit_state(state, action)
    onehot = [[1 if next_state == (row, col) else 0 for col in range(self.maze_shape[1])]\
        for row in range(self.maze_shape[0])]
    return flatten(onehot)

  def fetch_training_metric(self, metrics):
    return metrics

  def predict_value(self, model, X):
    return model.predict(np.array([X]))[0][0]




import math
from kyoka.value_function.base_table_action_value_function import BaseTableActionValueFunction

class MazeTableValueFunction(BaseTableActionValueFunction):

  def __init__(self, maze_shape):
    BaseTableActionValueFunction.__init__(self)
    self.maze_shape = maze_shape

  def generate_initial_table(self):
    height, width = self.maze_shape
    action_num = 4
    Q_table = [[[0 for a in range(action_num)] for j in range(width)] for i in range(height)]
    return Q_table

  def fetch_value_from_table(self, table, state, action):
    row, col = state
    Q_value = table[row][col][action]
    return Q_value

  def update_table(self, table, state, action, new_value):
    row, col = state
    table[row][col][action] = new_value
    return table




from kyoka.value_function.base_action_value_function import BaseActionValueFunction
from kyoka.value_function.base_state_value_function import BaseStateValueFunction

class BasePolicy(object):

  def __init__(self, domain, value_function):
    self.domain = domain
    self.value_function = value_function

  def choose_action(self, state, action=None):
    err_msg = self.__build_err_msg("choose_action")
    raise NotImplementedError(err_msg)

  def pack_arguments_for_value_function(self, state, action):
    if isinstance(self.value_function, BaseStateValueFunction):
      return (state)
    elif isinstance(self.value_function, BaseActionValueFunction):
      return (state, action)
    else:
      raise ValueError("Invalid value function is set")

  def __build_err_msg(self, msg):
    base_msg = "[ {0} ] class does not implement [ {1} ] method"
    return base_msg.format(self.__class__.__name__, msg)





import random
from kyoka.policy.base_policy import BasePolicy

class GreedyPolicy(BasePolicy):

  def __init__(self, domain, value_function, rand=None):
    super(GreedyPolicy, self).__init__(domain, value_function)
    self.rand = rand if rand else random

  def choose_action(self, state):
    actions = self.domain.generate_possible_actions(state)
    pack = lambda state, action: self.pack_arguments_for_value_function(state, action)
    calc_Q_value = lambda packed_arg: self.value_function.calculate_value(*packed_arg)
    Q_value_for_actions = [calc_Q_value(pack(state, action)) for action in actions]
    max_Q_value = max(Q_value_for_actions)
    Q_act_pair = zip(Q_value_for_actions, actions)
    best_actions = [act for Q_value, act in Q_act_pair if max_Q_value == Q_value]
    best_action = self.rand.choice(best_actions)
    return best_action




import os
from tests.base_unittest import BaseUnitTest
from kyoka.value_function.base_keras_action_value_function import BaseKerasActionValueFunction
from mock import Mock

class BaseKerasActionValueFunctionTest(BaseUnitTest):

  def setUp(self):
    self.func = self.TestImplementation()
    self.func.setUp()

  def test_calculate_value(self):
    state, action = 5, 7
    self.eq(36, self.func.calculate_value(state, action))

  def test_update_function(self):
    state, action, reward = 5, 7, 10
    delta = self.func.update_function(state, action, reward)
    self.func.model.train_on_batch.assert_called_with(35, reward)
    self.eq(2, delta)

    delta = self.func.update_function(state, action, reward)
    self.eq(0, delta)

  def test_save_model_weights(self):
    filepath = "hoge"
    self.func.save_model_weights(filepath)
    self.func.model.save_weights.assert_called_with(filepath)

  def test_load_model_weights(self):
    filepath = "hoge"
    self.func.load_model_weights(filepath)
    self.func.model.load_weights.assert_called_with(filepath)

  class TestImplementation(BaseKerasActionValueFunction):

    def generate_model(self):
      model = Mock()
      model.train_on_batch.return_value = [1, 2]
      return model

    def transform_state_action_into_input(self, state, action):
      return state * action

    def predict_value(self, model, X):
      return X+1

    def fetch_training_metric(self, metrics):
      return metrics[1]




from tests.base_unittest import BaseUnitTest
from sample.ticktacktoe.ticktacktoe_domain import TickTackToeDomain

class TickTackToeDomainTest(BaseUnitTest):
  """
  Board and bit-flg relation
  EX1.
    (O) first_player_board  = 110000000     OO-
    (X) second_player_board = 000000011  => ---
                                            -XX
  EX2.
    (O) first_player_board  = 101010101     OXO
    (X) second_player_board = 010101010  => XOX
                                            OXO
  """

  def setUp(self):
    self.domain = TickTackToeDomain()

  def test_generate_inital_state(self):
    state = self.domain.generate_initial_state()
    self.eq((0,0), state)

  def test_terminal_state_check_when_initial_state(self):
    state = self.__gen_initial_state()
    self.false(self.domain.is_terminal_state(state))

  def test_terminal_state_check_when_draw(self):
    state = self.__gen_draw_state()
    self.true(self.domain.is_terminal_state(state))

  def test_terminal_state_check_when_vertical_win(self):
    state = self.__gen_vertical_win_state()
    self.true(self.domain.is_terminal_state(state))

  def test_terminal_state_check_when_horizontal_win(self):
    state = self.__gen_horizontal_win_state()
    self.true(self.domain.is_terminal_state(state))

  def test_terminal_state_check_when_diagonal_win(self):
    state = self.__gen_diagonal_win_state()
    self.true(self.domain.is_terminal_state(state))

  def test_transit_state_from_initial_state(self):
    state = self.__gen_initial_state()
    state = self.domain.transit_state(state, 1)
    state = self.domain.transit_state(state, 2)
    state = self.domain.transit_state(state, 256)
    expected = (257, 2)
    self.eq(expected, state)

  def test_generate_possible_actions_when_draw(self):
    state = self.__gen_draw_state()
    expected = []
    self.eq(expected, self.domain.generate_possible_actions(state))

  def test_generate_possible_actions_when_initial_state(self):
    state = self.__gen_initial_state()
    expected = [1, 2, 4, 8, 16, 32, 64, 128, 256]
    self.eq(expected, self.domain.generate_possible_actions(state))

  def test_generate_possible_actions_when_vertical_win(self):
    state = self.__gen_vertical_win_state()
    expected = [1, 2, 8, 64]
    self.eq(expected, self.domain.generate_possible_actions(state))

  def test_generate_possible_actions_when_horizontal_win(self):
    state = self.__gen_horizontal_win_state()
    expected = [1, 2, 4, 8]
    self.eq(expected, self.domain.generate_possible_actions(state))

  def test_generate_possible_actions_when_diagonal_win(self):
    state = self.__gen_diagonal_win_state()
    expected = [2, 4, 8, 32]
    self.eq(expected, self.domain.generate_possible_actions(state))

  def test_calculate_reward_when_initial_state(self):
    state = self.__gen_initial_state()
    self.eq(TickTackToeDomain.REWARD_NONE, self.domain.calculate_reward(state))

  def test_calculate_reward_when_first_player_win(self):
    state = self.__gen_diagonal_win_state()
    self.eq(TickTackToeDomain.REWARD_WIN, self.domain.calculate_reward(state))

  def test_calculate_reward_when_first_player_lose(self):
    state = self.__gen_diagonal_win_state()[::-1]
    self.eq(TickTackToeDomain.REWARD_LOSE, self.domain.calculate_reward(state))

  def test_is_first_player_flg(self):
    domain = TickTackToeDomain(is_first_player=False)
    state = self.__gen_diagonal_win_state()
    self.eq(TickTackToeDomain.REWARD_LOSE, domain.calculate_reward(state))


  def __gen_initial_state(self):
    """
    ---
    ---
    ---
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("000000000")
    second_player_board = bin2i("000000000")
    state = (first_player_board, second_player_board)
    return state

  def __gen_draw_state(self):
    """
    OXO
    OXO
    XOX
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("010101101")
    second_player_board = bin2i("101010010")
    state = (first_player_board, second_player_board)
    return state

  def __gen_vertical_win_state(self):
    """
    OX-
    OX-
    O--
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("100100100")
    second_player_board = bin2i("010010000")
    state = (first_player_board, second_player_board)
    return state

  def __gen_horizontal_win_state(self):
    """
    OOO
    XX-
    ---
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("111000000")
    second_player_board = bin2i("000110000")
    state = (first_player_board, second_player_board)
    return state

  def __gen_diagonal_win_state(self):
    """
    OXX
    -O-
    --O
    """
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("100010001")
    second_player_board = bin2i("011000000")
    state = (first_player_board, second_player_board)
    return state




from kyoka.policy.base_policy import BasePolicy
import random
import logging

class TickTackToePerfectPolicy(BasePolicy):

  def choose_action(self, state):
    actions = self.domain.generate_possible_actions(state)
    states = [self.domain.transit_state(state, action) for action in actions]
    values = [self.mini(state, -20, 20) for state in states]
    logging.debug("MiniMax calculation result [(action, score),...] => %s" % zip(actions, values))
    best_actions = [act for act, val in zip(actions, values) if val == max(values)]
    return random.choice(best_actions)


  def maxi(self, state, alpha, beta):
    if self.domain.is_terminal_state(state):
      return self.domain.calculate_reward(state)
    for action in self.domain.generate_possible_actions(state):
      next_state = self.domain.transit_state(state, action)
      score = self.mini(next_state, alpha, beta)
      if score >= beta:
        return beta
      if score > alpha:
        alpha = score
    return alpha

  def mini(self, state, alpha, beta):
    if self.domain.is_terminal_state(state):
      return self.domain.calculate_reward(state)
    best = 100
    for action in self.domain.generate_possible_actions(state):
      next_state = self.domain.transit_state(state, action)
      score = self.maxi(next_state, alpha, beta)
      if score <= alpha:
        return alpha
      if score < beta:
        beta = score
    return beta




import math
from kyoka.domain.base_domain import BaseDomain

class TickTackToeDomain(BaseDomain):

  REWARD_WIN = 1
  REWARD_LOSE = -1
  REWARD_NONE = 0

  def __init__(self, is_first_player=True):
    self.is_first_player = is_first_player

  def generate_initial_state(self):
    first_player_board = 0
    second_player_board = 0
    state = (first_player_board, second_player_board)
    return state

  def is_terminal_state(self, state):
    return any(map(self.__is_winning, state) + [self.__is_draw(state)])

  def transit_state(self, state, action):
    first_player_board ,second_player_board = state
    count_move = lambda : bin(first_player_board | second_player_board).count("1")
    if count_move()%2==0:
      first_player_board |= action
    else:
      second_player_board |= action
    next_state = (first_player_board, second_player_board)
    return next_state

  def generate_possible_actions(self, state):
    board = state[0] | state[1]
    def add(ary, pos):
      if (board >> pos)&1 == 0:
        ary.append(1<<pos)
      return ary
    return reduce(add, range(9), [])

  def calculate_reward(self, state):
    first_player_board, second_player_board = state
    reward = self.REWARD_NONE
    if (self.__is_winning(first_player_board)):
      reward = self.REWARD_WIN
    elif(self.__is_winning(second_player_board)):
      reward = self.REWARD_LOSE
    return reward if self.is_first_player else -reward


  def __is_winning(self, player_board):
    bin2i = lambda b: int(b, 2)
    line_horizon = any([player_board & mask == mask for mask in map(bin2i, ['000000111', '000111000', '111000000'])])
    line_vertical = any([player_board & mask == mask for mask in map(bin2i, ['001001001', '010010010', '100100100'])])
    line_diagonal = any([player_board & mask == mask for mask in map(bin2i, ['100010001', '001010100'])])
    return line_horizon | line_vertical | line_diagonal

  def __is_draw(self, state):
    return len(self.generate_possible_actions(state)) == 0

  def __find_connected_line(self, player_board):
    bin2i = lambda b: int(b, 2)
    line_horizon = any([player_board & mask == mask for mask in map(bin2i, ['000000111', '000111000', '111000000'])])
    line_vertical = any([player_board & mask == mask for mask in map(bin2i, ['001001001', '010010010', '100100100'])])
    line_diagonal = any([player_board & mask == mask for mask in map(bin2i, ['100010001', '001010100'])])
    return line_horizon | line_vertical | line_diagonal




from tests.base_unittest import BaseUnitTest
from kyoka.finish_rule.watch_update_delta import WatchUpdateDelta

from nose.tools import raises

class WatchUpdateDeltaTest(BaseUnitTest):

  def test_satisfy_condition(self):
    rule = WatchUpdateDelta(patience=2, minimum_required_delta=2)
    self.false(rule.satisfy_condition("dummy", [1]))
    self.false(rule.satisfy_condition("dummy", [1, 2]))
    self.false(rule.satisfy_condition("dummy", [-1]))
    self.false(rule.satisfy_condition("dummy", [-1, -3]))
    self.false(rule.satisfy_condition("dummy", [1, -1]))
    self.true(rule.satisfy_condition("dummy", [1.9, -1]))

  def test_max_delta_memo_management(self):
    rule = WatchUpdateDelta(patience=2, minimum_required_delta=2)
    self.false(rule.satisfy_condition("dummy", [1.5]))
    self.include(1.5, rule.max_delta_memo)
    self.false(rule.satisfy_condition("dummy", [1, 2]))
    self.not_include(1.5, rule.max_delta_memo)


  def test_generate_progress_message(self):
    rule = WatchUpdateDelta(patience=2, minimum_required_delta=3)
    rule.satisfy_condition(1, [1.9, -1])
    msg = rule.generate_progress_message(5, "dummy")
    self.include(str(5), msg)
    self.include(str(1), msg)

  def test_generate_finish_message(self):
    rule = WatchUpdateDelta(patience=2, minimum_required_delta=3)
    rule.satisfy_condition(1, [1.9, -1])
    msg = rule.generate_finish_message(5, "dummy")
    self.include(str(2), msg)
    self.include(str(3), msg)
    self.include(str(1.9), msg)




from kyoka.policy.greedy_policy import GreedyPolicy

class MazeHelper:

  @classmethod
  def visualize_maze(self, maze):
    return "\n".join(maze)

  @classmethod
  def measure_performance(self, domain, value_function, step_limit=10000):
    policy = GreedyPolicy(domain, value_function)
    state = domain.generate_initial_state()
    step_counter = 0
    while not domain.is_terminal_state(state):
      action = policy.choose_action(state)
      state = domain.transit_state(state, action)
      step_counter += 1
      if step_counter >= step_limit:
        break
    return step_counter

  @classmethod
  def visualize_policy(self, domain, value_function):
    icon_map = { domain.UP: "^", domain.DOWN: "v", domain.RIGHT: ">", domain.LEFT: "<", -1: "-", -2: "G" }
    actions = self.__find_best_actions_on_each_cell(domain ,value_function)
    flg2icon = lambda flg: icon_map[flg]
    visualized_actions = [[flg2icon(flg) for flg in line] for line in actions]
    return self.visualize_maze(["".join(line) for line in visualized_actions])


  @classmethod
  def __find_best_actions_on_each_cell(self, domain, value_function):
    height, width = domain.get_maze_shape()
    goal_r, goal_c = domain.generate_terminal_state()
    curry = lambda row, col: self.__find_single_best_action(domain, value_function, row, col)
    maze_with_answer = [[curry(row, col) for col in range(width)] for row in range(height)]
    maze_with_answer[goal_r][goal_c] = -2
    return maze_with_answer

  @classmethod
  def __find_single_best_action(self, domain, value_function, row, col):
    state = (row, col)
    actions = domain.generate_possible_actions(state)
    values = [value_function.calculate_value(state, action) for action in actions]
    best_actions = [act for act, val in zip(actions, values) if val == max(values)]
    if len(best_actions) == 1:
      return best_actions[0]
    else:
      return -1




from tests.base_unittest import BaseUnitTest
from kyoka.value_function.base_state_value_function import BaseStateValueFunction

class BaseStateValueFunctionTest(BaseUnitTest):

  def setUp(self):
    self.func = BaseStateValueFunction()

  def test_error_msg_when_not_implement_abstract_method(self):
    try:
      self.func.calculate_value("dummy")
    except NotImplementedError as e:
      self.include("BaseStateValueFunction", str(e))
      self.include("calculate_value", str(e))
    else:
      self.fail("NotImplementedError does not occur")




from tests.base_unittest import BaseUnitTest
from sample.ticktacktoe.ticktacktoe_helper import TickTackToeHelper

class TickTackToeHelperTest(BaseUnitTest):

  def test_visualize_board_when_empty(self):
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("000000000")
    second_player_board = bin2i("000000000")
    state = (first_player_board, second_player_board)
    expected = "- - -\n- - -\n- - -"
    self.eq(expected, TickTackToeHelper.visualize_board(state))

  def test_visualize_board_when_not_empty(self):
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("110000000")
    second_player_board = bin2i("000000011")
    state = (first_player_board, second_player_board)
    expected = "O O -\n- - -\n- X X"
    self.eq(expected, TickTackToeHelper.visualize_board(state))




from tests.base_unittest import BaseUnitTest
from kyoka.algorithm.montecarlo.montecarlo import MonteCarlo
from kyoka.policy.greedy_policy import GreedyPolicy
from kyoka.value_function.base_action_value_function import BaseActionValueFunction
from kyoka.value_function.base_table_action_value_function import BaseTableActionValueFunction

from mock import Mock
from nose.tools import raises

class MonteCarloTest(BaseUnitTest):

  def setUp(self):
    self.algo = MonteCarlo()

  def test_update_value_function(self):
    domain = self.__setup_stub_domain()
    value_func = self.TestTableValueFunctionImpl()
    value_func.setUp()
    policy = GreedyPolicy(domain, value_func)
    delta = self.algo.update_value_function(domain, policy, value_func)
    self.eq([59, 58, 49], delta)
    expected = [(0, 1, 59, 1), (1, 2, 58, 1), (3, 4, 49, 1), (0, 0, 0, 0)]
    update_counter = value_func.get_additinal_data("additinal_data_key_montecarlo_update_counter")
    for state, action, value, update_count in expected:
      self.eq(value, value_func.fetch_value_from_table(value_func.table, state, action))
      self.eq(update_count, value_func.fetch_value_from_table(update_counter, state, action))

  def test_update_value_function_twice_for_counter(self):
    domain = self.__setup_stub_domain()
    value_func = self.TestTableValueFunctionImpl()
    value_func.setUp()
    policy = GreedyPolicy(domain, value_func)

    self.algo.update_value_function(domain, policy, value_func)
    domain.calculate_reward.side_effect = lambda state: state
    delta = self.algo.update_value_function(domain, policy, value_func)

    self.eq([-24, -24, -21], delta)
    expected = [(0, 1, 35, 2), (1, 2, 34, 2), (3, 4, 28, 2), (0, 0, 0, 0)]
    update_counter = value_func.get_additinal_data("additinal_data_key_montecarlo_update_counter")
    for state, action, value, update_count in expected:
      self.eq(value, value_func.fetch_value_from_table(value_func.table, state, action))
      self.eq(update_count, value_func.fetch_value_from_table(update_counter, state, action))

  def test_save_update_counter_as_additinal_data(self):
    domain = self.__setup_stub_domain()
    value_func = self.TestTableValueFunctionImpl()
    value_func.setUp()
    policy = GreedyPolicy(domain, value_func)
    delta = self.algo.update_value_function(domain, policy, value_func)
    update_counter = value_func.get_additinal_data("additinal_data_key_montecarlo_update_counter")
    self.eq(1, update_counter[0][1])
    self.eq(1, update_counter[1][2])
    self.eq(1, update_counter[3][4])
    self.eq(0, update_counter[0][0])

  @raises(TypeError)
  def test_table_value_function_validation(self):
    value_func = Mock(spec=BaseActionValueFunction)
    self.algo.update_value_function("dummy", "dummy", value_func)


  def __setup_stub_domain(self):
    mock_domain = Mock()
    mock_domain.generate_initial_state.return_value = 0
    mock_domain.is_terminal_state.side_effect = lambda state: state == 7
    mock_domain.transit_state.side_effect = lambda state, action: state + action
    mock_domain.generate_possible_actions.side_effect = lambda state: [state + 1]
    mock_domain.calculate_reward.side_effect = lambda state: state**2
    return mock_domain


  class TestTableValueFunctionImpl(BaseTableActionValueFunction):

    def generate_initial_table(self):
      return [[0 for j in range(50)] for i in range(4)]

    def fetch_value_from_table(self, table, state, action):
      return table[state][action]

    def update_table(self, table, state, action, new_value):
      table[state][action] = new_value
      return table





import os
from tests.base_unittest import BaseUnitTest
from sample.maze.maze_keras_value_function import MazeKerasValueFunction
from sample.maze.maze_domain import MazeDomain

class MazeKerasValueFunctionTest(BaseUnitTest):

  def setUp(self):
    self.domain = MazeDomain()
    self.domain.read_maze(self.__get_sample_maze_path())
    self.func = MazeKerasValueFunction(self.domain)
    self.func.setUp()

  def test_transform_state_action_into_input(self):
    test_func = lambda state, action: self.func.transform_state_action_into_input(state, action)
    state1, action1 = (0, 0), self.domain.RIGHT
    self.eq([0,1,0,0,0,0,0,0,0,0,0,0], test_func(state1, action1))
    state2, action2 = (2, 1), self.domain.UP
    self.eq([0,0,0,0,0,1,0,0,0,0,0,0], test_func(state2, action2))

  def _test_prediction(self):
    state, action = (0, 0), self.domain.DOWN
    delta = self.func.update_function(state, action, 1)
    res = self.func.calculate_value(state, action)
    self.debug()

  def __get_sample_maze_path(self):
    return os.path.join(os.path.dirname(__file__), "sample_maze.txt")




from kyoka.callback.base_callback import BaseCallback
import logging

class MazeTransformer(BaseCallback):

  def __init__(self):
    BaseCallback.__init__(self)
    self.transformation = {}

  def set_transformation(self, timing, maze_path):
    self.transformation[timing] = maze_path

  def before_update(self, iteration_count, domain, value_function):
    if iteration_count in self.transformation:
      maze_filepath = self.transformation[iteration_count]
      domain.read_maze(maze_filepath)
      logging.debug("Maze transformed into [ %s ]" % maze_filepath)




import os
from tests.base_unittest import BaseUnitTest
from kyoka.value_function.base_action_value_function import BaseActionValueFunction
from sample.maze.maze_domain import MazeDomain
from sample.maze.maze_helper import MazeHelper

class MazeHelperTest(BaseUnitTest):

  def setUp(self):
    self.domain = MazeDomain()
    self.domain.read_maze(self.__get_sample_maze_path())

  def test_visualize_maze(self):
    expected = \
        "-S--"+ "\n" + \
        "--XG"+ "\n" + \
        "----"
    self.eq(expected, MazeHelper.visualize_maze(self.domain.maze))

  def test_measure_performace(self):
    value_function = self.PerfectValueFunction()
    performance = MazeHelper.measure_performance(self.domain, value_function)
    self.eq(3, performance)

  def test_visualize_policy(self):
    value_function = self.PerfectValueFunction()
    expected = \
        "->>v"+ "\n" + \
        "---G"+ "\n" + \
        "----"
    self.eq(expected, MazeHelper.visualize_policy(self.domain, value_function))


  def test_find_best_actions_on_each_cell(self):
    value_function = self.PerfectValueFunction()
    answer = MazeHelper._MazeHelper__find_best_actions_on_each_cell(self.domain, value_function)
    self.eq(answer[0][0], -1)
    self.eq(answer[0][1], MazeDomain.RIGHT)
    self.eq(answer[0][2], MazeDomain.RIGHT)
    self.eq(answer[0][3], MazeDomain.DOWN)

  def __get_sample_maze_path(self):
    return os.path.join(os.path.dirname(__file__), "sample_maze.txt")

  class PerfectValueFunction(BaseActionValueFunction):

    CHEET_SHEET = {
        ((0,1), MazeDomain.RIGHT): 1,
        ((0,2), MazeDomain.RIGHT): 1,
        ((0,3), MazeDomain.DOWN): 1
    }

    def calculate_value(self, state, action):
      if (state, action) in self.CHEET_SHEET:
        return 1
      else:
        return 0




from tests.base_unittest import BaseUnitTest
from kyoka.algorithm.td_learning.q_learning import QLearning
from kyoka.value_function.base_table_action_value_function import BaseTableActionValueFunction
from kyoka.policy.base_policy import BasePolicy

from mock import Mock

class QLearningTest(BaseUnitTest):

  def setUp(self):
    self.algo = QLearning(alpha=0.5, gamma=0.1)

  def test_update_value_function(self):
    domain = self.__setup_stub_domain()
    value_func = self.TestTableValueFunctionImpl()
    value_func.setUp()
    value_func.update_function(1, 2, 10)
    value_func.update_function(1, 3, 11)
    value_func.update_function(3, 4, 100)
    value_func.update_function(3, 5, 101)
    policy = self.NegativePolicyImple(domain, value_func)
    delta = self.algo.update_value_function(domain, policy, value_func)
    self.almosteq([1.05, 4.55, -25.5], delta, tolerance=0.0001)
    expected = [(0, 1, 1.05), (1, 2, 14.55), (3, 4, 74.5)]
    for state, action, value in expected:
      self.eq(value, value_func.fetch_value_from_table(value_func.table, state, action))


  def __setup_stub_domain(self):
    mock_domain = Mock()
    mock_domain.generate_initial_state.return_value = 0
    mock_domain.is_terminal_state.side_effect = lambda state: state == 7
    mock_domain.transit_state.side_effect = lambda state, action: state + action
    mock_domain.generate_possible_actions.side_effect = lambda state: [] if state == 7 else [state + 1, state + 2]
    mock_domain.calculate_reward.side_effect = lambda state: state**2
    return mock_domain

  class TestTableValueFunctionImpl(BaseTableActionValueFunction):

    def generate_initial_table(self):
      return [[0 for j in range(50)] for i in range(8)]

    def fetch_value_from_table(self, table, state, action):
      try:
        table[state][action]
      except:
        from nose.tools import set_trace; set_trace()
      return table[state][action]

    def update_table(self, table, state, action, new_value):
      table[state][action] = new_value
      return table

  class NegativePolicyImple(BasePolicy):

    def choose_action(self, state):
      actions = self.domain.generate_possible_actions(state)
      calc_Q_value = lambda state, action: self.value_function.calculate_value(state, action)
      Q_value_for_actions = [calc_Q_value(state, action) for action in actions]
      min_Q_value = min(Q_value_for_actions)
      Q_act_pair = zip(Q_value_for_actions, actions)
      worst_actions = [act for Q_value, act in Q_act_pair if min_Q_value == Q_value]
      return worst_actions[0]




from tests.base_unittest import BaseUnitTest
from kyoka.policy.greedy_policy import GreedyPolicy
from kyoka.value_function.base_state_value_function import BaseStateValueFunction
from mock import Mock

class GreedyPolicyTest(BaseUnitTest):

  def test_choose_action_when_best_action_is_single(self):
    domain = self.__setup_domain_stub([1,2,3])
    value_func = self.__setup_value_function_stub([100, 50, 150])
    policy = GreedyPolicy(domain, value_func)
    greedy_action = policy.choose_action(state="dummy")
    self.eq(3, greedy_action)

  def test_choose_action_when_best_action_is_multiple(self):
    domain = self.__setup_domain_stub([1,2,3])
    value_func = self.__setup_value_function_stub([100, 50, 100])
    random = self.__setup_random()
    policy = GreedyPolicy(domain, value_func, rand=random)
    greedy_action = policy.choose_action(state="dummy")
    self.eq(3, greedy_action)

  def __setup_domain_stub(self, possible_actions):
    mock_domain = Mock()
    mock_domain.generate_possible_actions.return_value = possible_actions
    return mock_domain

  def __setup_value_function_stub(self, mock_return):
    mock_value_func = Mock(spec=BaseStateValueFunction)
    mock_value_func.calculate_value.side_effect = mock_return
    return mock_value_func

  def __setup_random(self):
    random = Mock()
    random.choice.side_effect = lambda ary: ary[1]
    return random




from tests.base_unittest import BaseUnitTest
from kyoka.algorithm.base_rl_algorithm import BaseRLAlgorithm
from kyoka.policy.greedy_policy import GreedyPolicy
from kyoka.value_function.base_action_value_function import BaseActionValueFunction

from mock import Mock

class BaseRLAlgorithmTest(BaseUnitTest):

  def setUp(self):
    self.algo = BaseRLAlgorithm()

  def test_error_msg_when_not_implement_abstract_method(self):
    self.__check_err_msg(lambda : self.algo.update_value_function("dummy", "dummy", "dummy"), "update_value_function")

  def test_gen_episode(self):
    domain = self.__setup_stub_domain()
    value_func = self.__setup_stub_value_function()
    policy = GreedyPolicy(domain, value_func)
    episode = self.algo.generate_episode(domain, policy)
    self.eq(3, len(episode))
    self.eq((0, 1, 1, 1), episode[0])
    self.eq((1, 2, 3, 9), episode[1])
    self.eq((3, 4, 7, 49), episode[2])

  def test_GPI(self):
    algo = self.TestImplementation()
    finish_rule = self.__setup_stub_finish_rule()
    finish_msg = algo.GPI("dummy", "dummy", "dummy", finish_rule)
    expected = (2, 2)
    self.eq(expected, finish_msg)

  def test_GPI_with_multiple_finish_rules(self):
    algo = self.TestImplementation()
    finish_rule1 = self.__setup_stub_finish_rule(satisfy_condition=[False, True])
    finish_rule2 = self.__setup_stub_finish_rule(satisfy_condition=[True, False])
    finish_rules = [finish_rule1, finish_rule2]
    finish_msg = algo.GPI("dummy", "dummy", "dummy", finish_rules)
    expected = (1, 1)
    self.eq(expected, finish_msg)

  def test_set_callback(self):
    algo = self.TestImplementation()
    callback = Mock()
    algo.set_gpi_callback(callback)
    finish_rule = self.__setup_stub_finish_rule()
    finish_msg = algo.GPI("domain", "dummy", "value_function", finish_rule)
    self.eq(1, callback.before_gpi_start.call_count)
    self.eq(2, callback.before_update.call_count)
    self.eq(2, callback.after_update.call_count)
    self.eq(1, callback.after_gpi_finish.call_count)
    callback.before_gpi_start.assert_called_with("domain", "value_function")
    callback.before_update.assert_called_with(1, "domain", "value_function")
    callback.after_update.assert_called_with(1, "domain", "value_function", 2)
    callback.after_gpi_finish.assert_called_with("domain", "value_function")

  def __setup_stub_domain(self):
    mock_domain = Mock()
    mock_domain.generate_initial_state.return_value = 0
    mock_domain.is_terminal_state.side_effect = lambda state: state == 7
    mock_domain.transit_state.side_effect = lambda state, action: state + action
    mock_domain.generate_possible_actions.side_effect = lambda state: [state + 1]
    mock_domain.calculate_reward.side_effect = lambda state: state**2
    return mock_domain

  def __setup_stub_value_function(self):
    mock_value_func = Mock(spec=BaseActionValueFunction)
    mock_value_func.calculate_value.return_value = 0
    return mock_value_func

  def __setup_stub_finish_rule(self, satisfy_condition=[False, True]):
    mock_finish_fule = Mock()
    mock_finish_fule.satisfy_condition.side_effect = satisfy_condition
    mock_finish_fule.generate_finish_message.side_effect = lambda counter, delta: (counter, delta)
    return mock_finish_fule

  def __check_err_msg(self, target_method, target_word):
    try:
      target_method()
    except NotImplementedError as e:
      self.include(target_word, str(e))
    else:
      self.fail("NotImplementedError does not occur")

  class TestImplementation(BaseRLAlgorithm):

    def __init__(self):
      BaseRLAlgorithm.__init__(self)
      self.update_count = 0

    def update_value_function(self, domain, policy, value_function):
      self.update_count += 1
      return self.update_count




from kyoka.domain.base_domain import BaseDomain

class MazeDomain(BaseDomain):

  UP    = 0
  DOWN  = 1
  RIGHT = 2
  LEFT  = 3

  def read_maze(self, filepath):
    with open(filepath, 'rb') as f:
      self.maze = [line.rstrip() for line in f.readlines()]
      self.validate_maze(self.maze)

  def get_maze_shape(self):
    width, height = len(self.maze[0]), len(self.maze)
    return (height, width)

  def generate_initial_state(self):
    return self.__find_cell('S')

  def is_terminal_state(self, state):
    row, col = state
    return 'G' == self.maze[row][col]

  def transit_state(self, state, action):
    row, col = state
    height, width = self.get_maze_shape()
    if action == self.UP:
      row = max(0, row-1)
    elif action == self.DOWN:
      row = min(height-1, row+1)
    elif action == self.RIGHT:
      col= min(width-1, col+1)
    elif action == self.LEFT:
      col = max(0, col-1)
    if 'X' != self.maze[row][col]:
      return (row, col)
    else:
      return state

  def generate_possible_actions(self, state):
    return [self.UP, self.DOWN, self.RIGHT, self.LEFT]

  def calculate_reward(self, state):
    row, col = state
    return 1 if 'G' == self.maze[row][col] else 0


  def validate_maze(self, maze):
    cells = reduce(lambda x, y: x+y, maze)
    start_count = cells.count('S')
    goal_count = cells.count('G')
    if start_count != 1 or goal_count != 1:
      raise ValueError("Invalid maze passed. reason: start_count=%d, goal_count=%d")

  def generate_terminal_state(self):
    return self.__find_cell('G')

  def __find_cell(self, target):
    for row in range(len(self.maze)):
      for col in range(len(self.maze[0])):
        if self.maze[row][col] == target:
          return (row, col)




from tests.base_unittest import BaseUnitTest
from kyoka.finish_rule.watch_iteration_count import WatchIterationCount

from nose.tools import raises

class BaseFinishRuleTest(BaseUnitTest):

  def setUp(self):
    self.rule = WatchIterationCount(target_count=100)

  def test_satisfy_condition(self):
    self.false(self.rule.satisfy_condition(99, "dummy"))
    self.true(self.rule.satisfy_condition(100, "dummy"))
    self.true(self.rule.satisfy_condition(101, "dummy"))

  def test_generate_progress_message(self):
    msg = self.rule.generate_progress_message(5, "dummy")
    self.include(str(5), msg)
    self.include(str(100), msg)

  def test_generate_finish_message(self):
    self.include(str(5), self.rule.generate_finish_message(5, "dummy"))




import pickle

class BaseValueFunction(object):

  __KEY_MAIN_DATA = "key_base_value_function_main_data"
  __KEY_ADDITINAL_DATA = "key_base_value_function_additinal_data"

  def __init__(self):
    self.additinal_data_holder = {}

  def setUp(self):
    pass

  def deepcopy(self):
    return self

  def save(self, save_file_path):
    data = {
        self.__KEY_MAIN_DATA: self.provide_data_to_store(),
        self.__KEY_ADDITINAL_DATA: self.additinal_data_holder
    }
    self.__pickle_data(save_file_path, data)

  def load(self, load_file_path):
    stored_data = self.__unpickle_data(load_file_path)
    self.receive_data_to_restore(stored_data[self.__KEY_MAIN_DATA])
    self.additinal_data_holder = stored_data[self.__KEY_ADDITINAL_DATA]

  def set_additinal_data(self, key, data):
    self.additinal_data_holder[key] = data

  def get_additinal_data(self, key):
    if self.additinal_data_holder.has_key(key):
      return self.additinal_data_holder[key]
    else:
      return None

  def provide_data_to_store(self):
    err_msg = self.__build_err_msg("provide_data_to_store")
    raise NotImplementedError(err_msg)

  def receive_data_to_restore(self, ):
    err_msg = self.__build_err_msg("receive_data_to_restore")
    raise NotImplementedError(err_msg)

  def __pickle_data(self, filepath, data):
    with open(filepath, "wb") as f:
      pickle.dump(data, f)

  def __unpickle_data(self, filepath):
    with open(filepath, "rb") as f:
      return pickle.load(f)

  def __build_err_msg(self, msg):
    base_msg = "save method called but [ {0} ] class does not implement [ {1} ] method"
    return base_msg.format(self.__class__.__name__, msg)




from kyoka.finish_rule.base_finish_rule import BaseFinishRule

class WatchIterationCount(BaseFinishRule):

  def __init__(self, target_count, log_interval=100):
    BaseFinishRule.__init__(self, log_interval)
    self.target_count = target_count

  def check_condition(self, iteration_count, _deltas):
    return iteration_count >= self.target_count

  def generate_progress_message(self, iteration_count, deltas):
    base_msg = "Finished %d / %d iterations"
    return base_msg % (iteration_count, self.target_count)

  def generate_finish_message(self, iteration_count, deltas):
    base_msg = "Completed GPI iteration for %d times."
    return base_msg % iteration_count

  def __build_err_msg(self, msg):
    base_msg = "[ {0} ] class does not implement [ {1} ] method"
    return base_msg.format(self.__class__.__name__, msg)




class BaseDomain:

  def __init__(self):
    pass

  def generate_inital_state(self):
    err_msg = self.__build_err_msg("generate_inital_state")
    raise NotImplementedError(err_msg)

  def is_terminal_state(self, state):
    err_msg = self.__build_err_msg("is_terminal_state")
    raise NotImplementedError(err_msg)

  def transit_state(self, state, action):
    err_msg = self.__build_err_msg("transit_state")
    raise NotImplementedError(err_msg)

  def generate_possible_actions(self, state):
    err_msg = self.__build_err_msg("generate_possible_actions")
    raise NotImplementedError(err_msg)

  def calculate_reward(self, state):
    err_msg = self.__build_err_msg("calculate_reward")
    raise NotImplementedError(err_msg)


  def __build_err_msg(self, msg):
    return "Your client does not implement [ {0} ] method".format(msg)




import math
from kyoka.value_function.base_table_action_value_function import BaseTableActionValueFunction

class TickTackToeTableValueFunction(BaseTableActionValueFunction):

  def generate_initial_table(self):
    board_state_num = 2**9
    action_num = 9
    Q_table = [[[0 for a in range(action_num)]\
        for j in range(board_state_num)] for i in range(board_state_num)]
    return Q_table

  def fetch_value_from_table(self, table, state, action):
    first_player_board, second_player_board = state
    move_position = int(math.log(action,2))
    Q_value = table[first_player_board][second_player_board][move_position]
    return Q_value

  def update_table(self, table, state, action, new_value):
    first_player_board, second_player_board = state
    move_position = int(math.log(action,2))
    table[first_player_board][second_player_board][move_position] = new_value
    return table




from sample.ticktacktoe.ticktacktoe_domain import TickTackToeDomain
from kyoka.value_function.base_keras_action_value_function import BaseKerasActionValueFunction
from keras.models import Sequential
from keras.layers.core import Dense, Activation
import numpy as np

class TickTackToeKerasValueFunction(BaseKerasActionValueFunction):

  def generate_model(self):
    model = Sequential()
    model.add(Dense(100, input_shape=(18,)))
    model.add(Activation("relu"))
    model.add(Dense(1))
    model.compile(loss="mse",  optimizer="adam")
    return model

  def transform_state_action_into_input(self, state, action):
    domain = TickTackToeDomain()
    next_state = domain.transit_state(state, action)
    flg_to_ary = lambda flg: reduce(lambda acc, e: acc + [1 if (flg>>e)&1==1 else 0], range(9), [])
    multi_dim_ary = [flg_to_ary(player_board) for player_board in next_state]
    return multi_dim_ary[0] + multi_dim_ary[1]

  def fetch_training_metric(self, metrics):
    return metrics

  def predict_value(self, model, X):
    return model.predict(np.array([X]))[0][0]




class TickTackToeHelper:

  @classmethod
  def visualize_board(self, state):
    visualize_move = lambda flg: {-1: "-", 0:"O", 1:"X"}[flg]
    fetch_line = lambda board, line_idx: board[line_idx*3:(line_idx+1)*3]
    visualize_line = lambda moves: "%s %s %s" % tuple(moves)

    board = self.__bit_to_array(state)
    visualized_moves = map(visualize_move, board)
    visualized_lines = [visualize_line(fetch_line(visualized_moves, i)) for i in range(3)]
    visualized_board = "\n".join(visualized_lines)
    return visualized_board


  @classmethod
  def __bit_to_array(self, state):
    board = [-1 for i in range(9)]
    for i in range(2):
      for j in range(9):
        if (state[i] >> j) & 1 == 1:
          board[j] = i
    return board[::-1]

  @classmethod
  def measure_performance(self, domain, players):
    next_is_first_player = lambda state: bin(state[0]|state[1]).count("1") % 2 == 0
    next_player = lambda state: players[0] if next_is_first_player(state) else players[1]
    state = domain.generate_initial_state()
    while not domain.is_terminal_state(state):
      action = next_player(state).choose_action(state)
      state = domain.transit_state(state, action)
    return domain.calculate_reward(state)




from kyoka.callback.base_callback import BaseCallback
from sample.maze.maze_helper import MazeHelper

class MazePerformanceLogger(BaseCallback):

  def before_gpi_start(self, domain, value_function):
    self.step_log = []
    self.policy_log = []

  def after_update(self, iteration_count, domain, value_function, delta):
    step_to_goal = MazeHelper.measure_performance(domain, value_function)
    self.step_log.append(step_to_goal)
    self.policy_log.append(MazeHelper.visualize_policy(domain, value_function))




import os
from kyoka.finish_rule.base_finish_rule import BaseFinishRule

class ManualInterruption(BaseFinishRule):

  TARGET_WARD = "stop"

  def __init__(self, monitor_file_path, log_interval=100):
    BaseFinishRule.__init__(self, log_interval)
    self.monitor_file_path = monitor_file_path

  def check_condition(self, _iteration_count, _deltas):
    return self.__order_found_in_monitoring_file(self.monitor_file_path, self.TARGET_WARD)

  def generate_progress_message(self, _iteration_count, _deltas):
    return None

  def generate_finish_message(self, iteration_count, _deltas):
    base_msg = "Interrupt GPI after %d iterations because interupption order found in [ %s ]."
    return base_msg % (iteration_count, self.monitor_file_path)

  def __order_found_in_monitoring_file(self, filepath, target_word):
    return os.path.isfile(filepath) and self.__found_target_ward_in_file(filepath, target_word)

  def __found_target_ward_in_file(self, filepath, target_word):
    search_word = lambda src, target: target in src
    src = self.__read_data(filepath)
    return search_word(src, target_word) if src else False

  def __read_data(self, filepath):
    with open(filepath, 'rb') as f:
      return f.read()




from tests.base_unittest import BaseUnitTest
from kyoka.domain.base_domain import BaseDomain
from kyoka.policy.base_policy import BasePolicy
from kyoka.value_function.base_action_value_function import BaseActionValueFunction
from kyoka.value_function.base_state_value_function import BaseStateValueFunction

from nose.tools import raises

class BasePolicyTest(BaseUnitTest):

  def setUp(self):
    domain = BaseDomain()
    value_func = BaseActionValueFunction()
    self.policy = BasePolicy(domain, value_func)

  def test_error_msg_when_not_implement_abstract_method(self):
    try:
      self.policy.choose_action(None, None)
    except NotImplementedError as e:
      self.include("BasePolicy", str(e))
    else:
      self.fail("NotImplementedError does not occur")

  def test_pack_arguments_for_value_function_when_action_value_function(self):
    policy = BasePolicy(BaseDomain(), BaseActionValueFunction())
    packed = policy.pack_arguments_for_value_function("state", "action")
    self.eq(("state", "action"), packed)

  def test_pack_arguments_for_value_function_when_state_value_function(self):
    policy = BasePolicy(BaseDomain(), BaseStateValueFunction())
    packed = policy.pack_arguments_for_value_function("state", "action")
    self.eq(("state"), packed)

  @raises(ValueError)
  def test_pack_arguments_for_value_function_when_illegal_state(self):
    policy = BasePolicy(BaseDomain(), "dummy")
    policy.pack_arguments_for_value_function("state", "action")




from kyoka.callback.base_callback import BaseCallback
from kyoka.policy.greedy_policy import GreedyPolicy
from sample.ticktacktoe.ticktacktoe_domain import TickTackToeDomain
from sample.ticktacktoe.ticktacktoe_table_value_function import TickTackToeTableValueFunction
from sample.ticktacktoe.ticktacktoe_perfect_policy import TickTackToePerfectPolicy
from sample.ticktacktoe.ticktacktoe_helper import TickTackToeHelper

class TickTackToePerformanceLogger(BaseCallback):

  def before_gpi_start(self, domain, value_function):
    self.game_log = []
    self.log_interval_counter = 0

  def after_update(self, iteration_count, domain, value_function, delta):
    iteration_count = iteration_count + 1
    if iteration_count % self.performance_test_interval == 0:
      players = self.__setup_players(value_function)
      game_results = [TickTackToeHelper.measure_performance(domain, players)\
        for _ in range(self.test_game_count)]
      result_count = [game_results.count(result) for result in [-1, 0, 1]]
      result_rate  = [1.0 * count / len(game_results) for count in result_count]
      self.game_log.append((iteration_count, result_rate))

  def set_performance_test_interval(self, interval):
      self.performance_test_interval = interval

  def set_test_game_count(self, count):
    self.test_game_count = count

  def set_is_first_player(self, is_first_player):
    self.is_first_player = is_first_player

  def __setup_players(self, value_function):
    domains = [TickTackToeDomain(is_first_player=is_first)\
        for is_first in [self.is_first_player, not self.is_first_player]]
    value_funcs = [value_function, TickTackToeTableValueFunction()]
    player_builders = [GreedyPolicy, TickTackToePerfectPolicy]
    players = [builder(domain, func) for builder, domain, func in\
        zip(player_builders, domains, value_funcs)]
    players = players if self.is_first_player else players[::-1]
    return players




from kyoka.value_function.base_action_value_function import BaseActionValueFunction

class BaseTableActionValueFunction(BaseActionValueFunction):

  def setUp(self):
    self.table = self.generate_initial_table()

  def calculate_value(self, state, action):
    return self.fetch_value_from_table(self.table, state, action)

  def update_function(self, state, action, new_value):
    old_value = self.fetch_value_from_table(self.table, state, action)
    delta = new_value - old_value
    self.update_table(self.table, state, action, new_value)
    return delta

  def provide_data_to_store(self):
    return self.table

  def receive_data_to_restore(self, restored_data):
    self.table = restored_data

  def generate_initial_table(self):
    err_msg = self.__build_err_msg("generate_initial_table")
    raise NotImplementedError(err_msg)

  def fetch_value_from_table(self, table, state, action):
    err_msg = self.__build_err_msg("fetch_value_from_table")
    raise NotImplementedError(err_msg)

  def update_table(self, table, state, action, new_value):
    err_msg = self.__build_err_msg("update_table")
    raise NotImplementedError(err_msg)


  def __build_err_msg(self, msg):
    base_msg = "[ {0} ] class does not implement [ {1} ] method"
    return base_msg.format(self.__class__.__name__, msg)




from tests.base_unittest import BaseUnitTest
from kyoka.policy.epsilon_greedy_policy import EpsilonGreedyPolicy
from kyoka.value_function.base_action_value_function import BaseActionValueFunction
from mock import Mock

class EpsilonGreedyPolicyTest(BaseUnitTest):

  def test_choose_action_boundary_test(self):
    eps = 0.0001
    self.eq(1, self.__choose_action_with_rand_value(0))
    self.eq(1, self.__choose_action_with_rand_value(0.1 - eps))
    self.eq(2, self.__choose_action_with_rand_value(0.1))
    self.eq(2, self.__choose_action_with_rand_value(0.2 + (1-0.3) - eps))
    self.eq(3, self.__choose_action_with_rand_value(0.2 + (1-0.3)))
    self.eq(3, self.__choose_action_with_rand_value(1-eps))

  def xtest_sampling_action(self):
    domain = self.__setup_domain_stub([1,2,3])
    value_func = self.__setup_value_function_stub([100, 150, 50]*100000)
    policy = EpsilonGreedyPolicy(domain, value_func, eps=0.3)
    result = [policy.choose_action(state="dummy") for _ in range(100000)]
    count = (result.count(1), result.count(2), result.count(3))
    self.debug()
    # sample result => count = (10062, 79955, 9983)


  def __choose_action_with_rand_value(self, rand_val):
    domain = self.__setup_domain_stub([1,2,3])
    value_func = self.__setup_value_function_stub([100, 150, 50])
    random = self.__setup_random(rand_val)
    policy = EpsilonGreedyPolicy(domain, value_func, eps=0.3, rand=random)
    return policy.choose_action(state="dummy")


  def __setup_domain_stub(self, possible_actions):
    mock_domain = Mock()
    mock_domain.generate_possible_actions.return_value = possible_actions
    return mock_domain

  def __setup_value_function_stub(self, mock_return):
    mock_value_func = Mock(spec=BaseActionValueFunction)
    mock_value_func.calculate_value.side_effect = mock_return
    return mock_value_func

  def __setup_random(self, rand_val):
    random = Mock()
    random.random.return_value = rand_val
    random.choice.side_effect = lambda ary: ary[0]
    return random




import logging

class BaseFinishRule(object):

  def __init__(self, log_interval=100):
    self.log_interval = log_interval
    self.log_interval_counter = 0

  def satisfy_condition(self, iteration_count, deltas):
    is_satisfied = self.check_condition(iteration_count, deltas)
    self.__notify_message_if_needed(is_satisfied, iteration_count, deltas)
    return is_satisfied

  def check_condition(self, iteration_count, deltas):
    err_msg = self.__build_err_msg("check_condition")
    raise NotImplementedError(err_msg)

  def generate_progress_message(self, iteration_count, deltas):
    err_msg = self.__build_err_msg("generate_progress_message")
    raise NotImplementedError(err_msg)

  def generate_finish_message(self, iteration_count, deltas):
    err_msg = self.__build_err_msg("generate_finish_message")
    raise NotImplementedError(err_msg)

  def __notify_message_if_needed(self, is_satisfied_condition, iteration_count, deltas):
    self.log_interval_counter += 1
    if is_satisfied_condition:
      self.__log(self.generate_finish_message(iteration_count, deltas))
    elif self.log_interval_counter >= self.log_interval:
      self.__log(self.generate_progress_message(iteration_count, deltas))
      self.log_interval_counter = 0

  def __build_err_msg(self, msg):
    base_msg = "[ {0} ] class does not implement [ {1} ] method"
    return base_msg.format(self.__class__.__name__, msg)

  def __log(self, message):
    if message and len(message) != 0:
      logging.info(message)




from kyoka.policy.base_policy import BasePolicy

class TickTackToeManualPolicy(BasePolicy):

  ACTION_NAME_MAP = {
        1 : "lower_right",
        2 : "lower_center",
        4 : "lower_left",
        8 : "middle_right",
        16: "middle_center",
        32: "middle_left",
        64: "upper_right",
        128: "upper_center",
        256: "upper_left"
  }

  def choose_action(self, state):
    message = self.__ask_message(state) + " >> "
    action = int(raw_input(message))
    if action not in self.domain.generate_possible_actions(state):
      return self.choose_action(state)
    return action

  def __ask_message(self, state):
    possible_actions = self.domain.generate_possible_actions(state)
    names = [self.ACTION_NAME_MAP[action] for action in possible_actions]
    return ", ".join(["%d: %s" % info for info in zip(possible_actions, names)])




import os
from tests.base_unittest import BaseUnitTest
from kyoka.value_function.base_value_function import BaseValueFunction
from kyoka.value_function.base_action_value_function import BaseActionValueFunction
from kyoka.value_function.base_state_value_function import BaseStateValueFunction

class BaseValueFunctionTest(BaseUnitTest):

  def tearDown(self):
    file_path = self.__generate_tmp_file_path()
    if os.path.isfile(file_path):
      os.remove(file_path)

  def test_deepcopy_default_implementation(self):
    func = BaseValueFunction()
    func.tmp = "hoge"
    copy = func.deepcopy()
    copy.tmp = "fuga"
    self.eq(func.tmp, copy.tmp)

  def test_save_load_function_with_action_value_function(self):
    file_path = self.__generate_tmp_file_path()
    original = self.TestActionValueFunction()
    original.state = "fuga"
    original.save(file_path)
    restored = self.TestActionValueFunction()
    self.eq("hoge", restored.state)
    restored.load(file_path)
    self.eq("fuga", restored.state)

  def test_save_load_function_with_state_value_function(self):
    file_path = self.__generate_tmp_file_path()
    original = self.TestStateValueFunction()
    original.state = "fuga"
    original.save(file_path)
    restored = self.TestStateValueFunction()
    self.eq("hoge", restored.state)
    restored.load(file_path)
    self.eq("fuga", restored.state)

  def test_save_load_additional_data(self):
    unique_key = "key4base_value_function_test"
    file_path = self.__generate_tmp_file_path()
    original = self.TestActionValueFunction()
    self.eq(None, original.get_additinal_data(unique_key))
    original.set_additinal_data(unique_key, "hoge")
    self.eq("hoge", original.get_additinal_data(unique_key))
    original.save(file_path)
    restored = self.TestActionValueFunction()
    self.eq(None, restored.get_additinal_data(unique_key))
    restored.load(file_path)
    self.eq("hoge", restored.get_additinal_data(unique_key))


  def __generate_tmp_file_path(self):
    return os.path.join(os.path.dirname(__file__), "tmp_file_for_base_value_function_test.tmp")

  class TestActionValueFunction(BaseActionValueFunction):

    def __init__(self):
      BaseActionValueFunction.__init__(self)
      self.state = "hoge"

    def provide_data_to_store(self):
      return self.state

    def receive_data_to_restore(self, restored_data):
      self.state = restored_data

  class TestStateValueFunction(BaseStateValueFunction):

    def __init__(self):
      BaseStateValueFunction.__init__(self)
      self.state = "hoge"

    def provide_data_to_store(self):
      return self.state

    def receive_data_to_restore(self, restored_data):
      self.state = restored_data




from kyoka.algorithm.base_rl_algorithm import BaseRLAlgorithm
from kyoka.policy.greedy_policy import GreedyPolicy

class QLearning(BaseRLAlgorithm):

  ACTION_ON_TERMINAL_FLG = "action_on_terminal"

  def __init__(self, alpha=0.1, gamma=0.9):
    BaseRLAlgorithm.__init__(self)
    self.alpha = alpha
    self.gamma = gamma

  def update_value_function(self, domain, policy, value_function):
    greedy_policy = GreedyPolicy(domain, value_function)
    state = domain.generate_initial_state()
    action = policy.choose_action(state)
    delta_history = []
    while not domain.is_terminal_state(state):
      next_state = domain.transit_state(state, action)
      reward = domain.calculate_reward(next_state)
      next_action = self.__choose_action(domain, policy, next_state)
      greedy_action = self.__choose_action(domain, greedy_policy, next_state)
      new_Q_value = self.__calculate_new_Q_value(\
          value_function, state, action, next_state, greedy_action, reward)
      delta = value_function.update_function(state, action, new_Q_value)
      delta_history.append(delta)
      state, action = next_state, next_action
    return delta_history


  def __calculate_new_Q_value(self,\
      value_function, state, action, next_state, greedy_action, reward):
    Q_value = value_function.calculate_value(state, action)
    greedy_Q_value = self.__calculate_value(value_function, next_state, greedy_action)
    return Q_value + self.alpha * (reward + self.gamma * greedy_Q_value - Q_value)

  def __choose_action(self, domain, policy, state):
    if domain.is_terminal_state(state):
      return self.ACTION_ON_TERMINAL_FLG
    else:
      return policy.choose_action(state)

  def __calculate_value(self, value_function, next_state, next_action):
    if self.ACTION_ON_TERMINAL_FLG == next_action:
      return 0
    else:
      return value_function.calculate_value(next_state, next_action)




import os
from tests.base_unittest import BaseUnitTest
from kyoka.value_function.base_table_state_value_function import BaseTableStateValueFunction

class BaseTableStateValueFunctionTest(BaseUnitTest):

  def setUp(self):
    self.func = self.TestImplementation()
    self.func.setUp()

  def tearDown(self):
    file_path = self.__generate_tmp_file_path()
    if os.path.isfile(file_path):
      os.remove(file_path)

  def test_calculate_value(self):
    state = 2
    self.func.table[state] = 1
    self.eq(1, self.func.calculate_value(state))

  def test_update_function(self):
    state = 2
    self.eq(0, self.func.calculate_value(state))

    delta = self.func.update_function(state, 1)
    self.eq(1, self.func.calculate_value(state))
    self.eq(1, delta)

    delta = self.func.update_function(state, 0)
    self.eq(0, self.func.calculate_value(state))
    self.eq(-1, delta)

  def test_store_and_restore_table(self):
    file_path = self.__generate_tmp_file_path()
    state = 0
    self.func.update_function(state, 1)
    self.eq(1, self.func.calculate_value(state))
    self.func.save(file_path)
    self.func = self.TestImplementation()
    self.func.load(file_path)
    self.eq(1, self.func.calculate_value(state))


  def __generate_tmp_file_path(self):
    return os.path.join(os.path.dirname(__file__), "tmp_file_for_base_table_state_value_function_test.tmp")

  class TestImplementation(BaseTableStateValueFunction):

    def generate_initial_table(self):
      return [0 for i in range(3)]

    def fetch_value_from_table(self, table, state):
      return table[state]

    def update_table(self, table, state, new_value):
      table[state] = new_value
      return table




from kyoka.algorithm.base_rl_algorithm import BaseRLAlgorithm
from kyoka.value_function.base_table_action_value_function import BaseTableActionValueFunction
from kyoka.value_function.base_table_state_value_function import BaseTableStateValueFunction

class MonteCarlo(BaseRLAlgorithm):

  __KEY_ADDITIONAL_DATA = "additinal_data_key_montecarlo_update_counter"

  def update_value_function(self, domain, policy, value_function):
    self.__validate_value_function(value_function)
    self.__initialize_update_counter_if_needed(value_function)
    update_counter = value_function.get_additinal_data(self.__KEY_ADDITIONAL_DATA)
    episode = self.generate_episode(domain, policy)
    delta_history = []
    for idx, turn_info in enumerate(episode):
      state, action, _next_state, _reward = turn_info
      Q_value = value_function.calculate_value(state, action)
      update_count = value_function.fetch_value_from_table(update_counter, state, action)
      following_reward = self.__calculate_following_state_reward(idx, episode)
      new_Q_value = self.__calculate_new_Q_value(Q_value, update_count, following_reward)
      value_function.update_table(update_counter, state, action, update_count + 1)
      delta = value_function.update_function(state, action, new_Q_value)
      delta_history.append(delta)
    value_function.set_additinal_data(self.__KEY_ADDITIONAL_DATA, update_counter)
    return delta_history

  def __validate_value_function(self, value_function):
    valid_type = isinstance(value_function, BaseTableActionValueFunction) or \
        isinstance(value_function, BaseTableStateValueFunction)
    if not valid_type:
      raise TypeError(self.__build_type_error_message())

  def __build_type_error_message(self):
    return 'MonteCarlo method requires you to use "table" type action-value function.\
        (child class of [BaseTableActionValueFunction])'

  def __initialize_update_counter_if_needed(self, value_function):
    if value_function.get_additinal_data(self.__KEY_ADDITIONAL_DATA) is None:
      update_counter = value_function.generate_initial_table()
      value_function.set_additinal_data(self.__KEY_ADDITIONAL_DATA, update_counter)

  def __calculate_following_state_reward(self, current_turn, episode):
    following_turn_info = episode[current_turn:]
    following_rewards = [reward for _, _, _, reward in following_turn_info]
    return sum(following_rewards)

  def __calculate_new_Q_value(self, Q_val_average, update_count, update_reward):
    return self.__calc_average_in_incremental_way(update_count, update_reward, Q_val_average)

  def __calc_average_in_incremental_way(self, k, r, Q):
    return Q + 1.0 / (k + 1) * (r - Q)




from collections import defaultdict
from tests.base_unittest import BaseUnitTest
from kyoka.algorithm.td_learning.eligibility_trace.action_eligibility_trace import ActionEligibilityTrace

from nose.tools import raises

class ActionEligibilityTraceTest(BaseUnitTest):

  def test_default_value(self):
    trace = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    self.eq(0, trace.get(0, 0))
    self.eq(0, trace.get(100, 200))

  def test_accumulating_update(self):
    trace = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    trace.update(100, 200)
    self.eq(1, trace.get(100, 200))
    trace.update(100, 200)
    self.eq(2, trace.get(100, 200))

  def test_replacing_update(self):
    trace = ActionEligibilityTrace("replacing_trace", gamma=0.5, lambda_=0.1)
    trace.update(100, 200)
    self.eq(1, trace.get(100, 200))
    trace.update(100, 200)
    self.eq(1, trace.get(100, 200))

  def test_decay(self):
    trace = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    trace.update(100, 200)
    trace.decay(100, 200)
    self.eq(0.05, trace.get(100, 200))

  def test_decay_discard(self):
    trace = ActionEligibilityTrace(\
        "accumulating_trace", discard_threshold=0.0001, gamma=0.99, lambda_=0.9)
    trace.update(0, 0)
    trace.update(100, 200)
    [trace.decay(100, 200) for _ in range(79)]
    self.neq(0, trace.get(100, 200))
    trace.decay(100, 200)
    self.eq([(0, 0, 1)], trace.get_eligibilities())

  def test_get_eligibilities(self):
    trace = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    trace.update(0, 0)
    trace.update(0, 0)
    trace.update(100, 200)
    expected = { 0 : { 0 : 2 }, 100: { 200: 1 } }
    eligibilities = trace.get_eligibilities()
    self.eq(2, len(eligibilities))
    for state, action, eligibility in eligibilities:
      self.eq(expected[state][action], eligibility)

  def test_clear(self):
    trace = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    trace.update(0, 0)
    trace.update(100, 200)
    trace.clear()
    self.eq(0, trace.get(0, 0))
    self.eq(0, trace.get(100, 2000))

  def test_dump_load(self):
    trace1 = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    trace1.update(0, 0)
    trace1.update(0, 0)
    trace1.update(100, 200)
    dump = trace1.dump()
    trace2 = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    trace2.load(dump)
    expected = { 0 : { 0 : 2 }, 100: { 200: 1 } }
    eligibilities = trace2.get_eligibilities()
    self.eq(2, len(eligibilities))
    for state, action, eligibility in eligibilities:
      self.eq(expected[state][action], eligibility)

  @raises(TypeError)
  def test_update_type_validation_when_initialize(self):
    ActionEligibilityTrace("invalid_type", gamma=0.5, lambda_=0.1)

  @raises(TypeError)
  def test_update_type_validation_when_update(self):
    trace = ActionEligibilityTrace("accumulating_trace", gamma=0.5, lambda_=0.1)
    trace.update_type = "invalid_type"
    trace.update(0, 0)




import random
from kyoka.policy.base_policy import BasePolicy

class EpsilonGreedyPolicy(BasePolicy):

  def __init__(self, domain, value_function, eps=0.05, rand=None):
    super(EpsilonGreedyPolicy, self).__init__(domain, value_function)
    self.eps = eps
    self.rand = rand if rand else random

  def choose_action(self, state):
    actions = self.domain.generate_possible_actions(state)
    best_action = self.__choose_best_action(state)
    probs = self.__calc_select_probability(best_action, actions)
    selected_action_idx = self.__roulette(probs)
    return actions[selected_action_idx]

  def __choose_best_action(self, state):
    actions = self.domain.generate_possible_actions(state)
    pack = lambda state, action: self.pack_arguments_for_value_function(state, action)
    calc_Q_value = lambda packed_arg: self.value_function.calculate_value(*packed_arg)
    Q_value_for_actions = [calc_Q_value(pack(state, action)) for action in actions]
    max_Q_value = max(Q_value_for_actions)
    Q_act_pair = zip(Q_value_for_actions, actions)
    best_actions = [act for Q_value, act in Q_act_pair if max_Q_value == Q_value]
    best_action = self.rand.choice(best_actions)
    return best_action

  def __calc_select_probability(self, best_action, actions):
    e = self.eps / len(actions)
    bonus = 1 - self.eps
    calc_prob = lambda action: e + bonus if action == best_action else e
    return [calc_prob(action) for action in actions]

  def __roulette(self, probs):
    hit_idx = -1
    prob_sum = 0
    dart = self.rand.random()
    for idx, prob in enumerate(probs):
      prob_sum += prob
      if dart < prob_sum:
        hit_idx = idx
        break
    return hit_idx




class BaseCallback(object):

  def before_gpi_start(self, domain, value_function):
    pass

  def before_update(self, iteration_count, domain, value_function):
    pass

  def after_update(self, iteration_count, domain, value_function, delta):
    pass

  def after_gpi_finish(self, domain, value_function):
    pass




from kyoka.value_function.base_value_function import BaseValueFunction

class BaseActionValueFunction(BaseValueFunction):

  def calculate_value(self, state, action):
    err_msg = self.__build_err_msg("calculate_value")
    raise NotImplementedError(err_msg)

  def update_function(self, state, action, new_value):
    err_msg = self.__build_err_msg("update_function")
    raise NotImplementedError(err_msg)

  def setUp(self):
    pass

  def provide_data_to_store(self):
    return None

  def receive_data_to_restore(self, restored_data):
    pass


  def __build_err_msg(self, msg):
    base_msg = "[ {0} ] class does not implement [ {1} ] method"
    return base_msg.format(self.__class__.__name__, msg)




from tests.base_unittest import BaseUnitTest
from sample.ticktacktoe.ticktacktoe_keras_value_function import TickTackToeKerasValueFunction

class TickTackToeKerasValueFunctionTest(BaseUnitTest):

  def setUp(self):
    self.func = TickTackToeKerasValueFunction()
    self.func.setUp()

  def test_state_action_into_input(self):
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("000000100")
    second_player_board = bin2i("000000010")
    action = 1
    model_input = self.func.transform_state_action_into_input((first_player_board, second_player_board), action)
    self.eq([1,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0], model_input)

  def _test_prediction(self):
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("000000100")
    second_player_board = bin2i("000000010")
    state = (first_player_board, second_player_board)
    action = 1
    model_input = self.func.transform_state_action_into_input((first_player_board, second_player_board), action)
    delta = self.func.update_function(state, action, 0)
    res = self.func.calculate_value(state, action)
    self.debug()




from collections import defaultdict

class ActionEligibilityTrace:

  TYPE_ACCUMULATING = "accumulating_trace"
  TYPE_REPLACING = "replacing_trace"

  def __init__(self, update_type, discard_threshold=0.0001, gamma=0.99, lambda_=0.9):
    self.__validate_update_type(update_type)
    self.update_type = update_type
    self.discard_threshold = discard_threshold
    self.gamma = gamma
    self.lambda_ = lambda_
    self.eligibility_holder = self.__generate_action_eligibility_holder()

  def get(self, state, action):
    return self.eligibility_holder[state][action]

  def update(self, state, action):
    if self.TYPE_ACCUMULATING == self.update_type:
      self.__update(state, action, self.get(state, action) + 1)
    elif self.TYPE_REPLACING == self.update_type:
      self.__update(state, action, 1)
    else:
      self.__validate_update_type(self.update_type)

  def decay(self, state, action):
    decayed = self.gamma * self.lambda_ * self.get(state, action)
    self.__update(state, action, decayed)
    if self.get(state, action) <= self.discard_threshold:
      self.__discard(state, action)

  def get_eligibilities(self):
    eligibilities = []
    for state in self.eligibility_holder:
      for action in self.eligibility_holder[state]:
        eligibility = self.eligibility_holder[state][action]
        eligibilities.append((state, action, eligibility))
    return eligibilities

  def clear(self):
    self.eligibility_holder = self.__generate_action_eligibility_holder()

  def dump(self):
    return self.get_eligibilities()

  def load(self, serial):
    self.clear()
    for state, action, eligibility in serial:
      self.__update(state, action, eligibility)

  def __validate_update_type(self, update_type):
    if not update_type in [self.TYPE_ACCUMULATING, self.TYPE_REPLACING]:
      raise TypeError("unknown update type [ %s ] passed" % update_type)

  def __update(self, state, action, new_value):
    self.eligibility_holder[state][action] = new_value

  def __discard(self, state, action):
    return self.eligibility_holder[state].pop(action)

  def __generate_action_eligibility_holder(self):
    return defaultdict(lambda: defaultdict(float))




class BaseRLAlgorithm(object):

  def __init__(self):
    self.callbacks = []

  def update_value_function(self, domain, policy, value_function):
    err_msg = self.__build_err_msg("update_value_function")
    raise NotImplementedError(err_msg)

  def GPI(self, domain, policy, value_function, finish_rules, debug=False):
    iteration_counter = 0
    [callback.before_gpi_start(domain, value_function) for callback in self.callbacks]
    while True:
      [callback.before_update(iteration_counter, domain, value_function) for callback in self.callbacks]
      delta = self.update_value_function(domain, policy, value_function)
      [callback.after_update(iteration_counter, domain, value_function, delta) for callback in self.callbacks]
      iteration_counter += 1
      for finish_rule in self.__wrap_rule_if_single(finish_rules):
        if finish_rule.satisfy_condition(iteration_counter, delta):
          finish_msg = finish_rule.generate_finish_message(iteration_counter, delta)
          [callback.after_gpi_finish(domain, value_function) for callback in self.callbacks]
          return finish_msg

  def generate_episode(self, domain, policy):
    state = domain.generate_initial_state()
    episode = []
    while not domain.is_terminal_state(state):
      action = policy.choose_action(state)
      next_state = domain.transit_state(state, action)
      reward = domain.calculate_reward(next_state)
      episode.append((state, action, next_state, reward))
      state = next_state
    return episode

  def set_gpi_callback(self, callback):
    self.callbacks.append(callback)


  def __wrap_rule_if_single(self, finish_rule):
    return [finish_rule] if not isinstance(finish_rule, list) else finish_rule

  def __build_err_msg(self, msg):
    return "Accessed [ {0} ] method of BaseRLAlgorithm which should be overridden".format(msg)




from tests.base_unittest import BaseUnitTest
from kyoka.finish_rule.base_finish_rule import BaseFinishRule

from nose.tools import raises
from mock import patch

class BaseFinishRuleTest(BaseUnitTest):

  def setUp(self):
    self.rule = BaseFinishRule()

  def test_satisfy_condition_logging(self):
    rule = self.TestImplementation(log_interval=2)
    with patch("logging.info") as log_mock:
      self.false(rule.satisfy_condition(1, 0))
      log_mock.assert_not_called()
      self.false(rule.satisfy_condition(2, 0))
      log_mock.assert_called_with("progress:2:0")
      self.false(rule.satisfy_condition(3, 0))
      log_mock.assert_called_with("progress:2:0")
      self.true(rule.satisfy_condition(0, 0))
      log_mock.assert_called_with("finish:0:0")

  @raises(NotImplementedError)
  def test_check_condition(self):
    self.rule.check_condition("dummy", "dummy")

  @raises(NotImplementedError)
  def test_generate_progress_condition(self):
    self.rule.check_condition("dummy", "dummy")

  @raises(NotImplementedError)
  def test_generate_finish_message(self):
    self.rule.generate_finish_message("dummy", "dummy")

  class TestImplementation(BaseFinishRule):

    def check_condition(self, iteration_count, deltas):
      return iteration_count == 0

    def generate_progress_message(self, iteration_count, deltas):
      return "%s:%s:%s" % ("progress", iteration_count, deltas)

    def generate_finish_message(self, iteration_count, deltas):
      return "%s:%s:%s" % ("finish", iteration_count, deltas)




class DomainValidator:

  def __init__(self, target_domain):
    self.domain = target_domain

  def implementation_check(self):
    def check(method, arg):
      try:
        method(*arg)
      except Exception as e:
        if isinstance(e, NotImplementedError):
          return e

    method_names = [
        ("generate_inital_state", 0),
        ("is_terminal_state", 1),
        ("transit_state", 2),
        ("generate_possible_actions", 1),
        ("calculate_reward", 1),
    ]
    methods = [(getattr(self.domain, method_name), range(arg_num))\
        for method_name, arg_num in method_names]
    errors = [check(method, arg) for method, arg in methods]
    message = "\n".join([str(err) for err in errors if err is not None])
    return len(errors)==0, message





from kyoka.algorithm.base_rl_algorithm import BaseRLAlgorithm
from kyoka.algorithm.td_learning.eligibility_trace.action_eligibility_trace\
    import ActionEligibilityTrace as EligibilityTrace

class SarsaLambda(BaseRLAlgorithm):

  __KEY_ADDITIONAL_DATA = "additinal_data_key_sarsa_lambda_eligibility_trace"
  ACTION_ON_TERMINAL_FLG = "action_on_terminal"

  def __init__(self, alpha=0.1, gamma=0.9, eligibility_trace=None):
    BaseRLAlgorithm.__init__(self)
    self.alpha = alpha
    self.gamma = gamma
    self.trace = eligibility_trace if eligibility_trace else self.__generate_default_trace()

  def update_value_function(self, domain, policy, value_function):
    self.__setup_trace(value_function)
    current_state = domain.generate_initial_state()
    current_action = policy.choose_action(current_state)
    update_delta_history = []
    while not domain.is_terminal_state(current_state):
      next_state = domain.transit_state(current_state, current_action)
      reward = domain.calculate_reward(next_state)
      next_action = self.__choose_action(domain, policy, next_state)
      delta = self.__calculate_delta(\
          value_function, current_state, current_action, next_state, next_action, reward)
      self.trace.update(current_state, current_action)
      for state, action, eligibility in self.trace.get_eligibilities():
        new_Q_value = self.__calculate_new_Q_value(\
            value_function, state, action, eligibility, delta)
        update_delta = value_function.update_function(state, action, new_Q_value)
        update_delta_history.append(update_delta)
        self.trace.decay(state, action)
      current_state, current_action = next_state, next_action
    self.__save_trace(value_function)
    return update_delta_history


  def __calculate_delta(self,\
      value_function, state, action,next_state, next_action, reward):
    Q_value = value_function.calculate_value(state, action)
    next_Q_value = self.__calculate_value(value_function, next_state, next_action)
    return reward + self.gamma * next_Q_value - Q_value

  def __calculate_new_Q_value(self, value_function, state, action, eligibility, delta):
    Q_value = self.__calculate_value(value_function, state, action)
    return Q_value + self.alpha * delta * eligibility

  def __choose_action(self, domain, policy, state):
    if domain.is_terminal_state(state):
      return self.ACTION_ON_TERMINAL_FLG
    else:
      return policy.choose_action(state)

  def __calculate_value(self, value_function, next_state, next_action):
    if self.ACTION_ON_TERMINAL_FLG == next_action:
      return 0
    else:
      return value_function.calculate_value(next_state, next_action)

  def __generate_default_trace(self):
    return EligibilityTrace(EligibilityTrace.TYPE_ACCUMULATING)

  def __setup_trace(self, value_function):
    trace_dump = value_function.get_additinal_data(self.__KEY_ADDITIONAL_DATA)
    if trace_dump:
      self.trace.load(trace_dump)

  def __save_trace(self, value_function):
    value_function.set_additinal_data(self.__KEY_ADDITIONAL_DATA, self.trace.dump())




from tests.base_unittest import BaseUnitTest
from kyoka.domain.base_domain import BaseDomain
from kyoka.domain.domain_validator import DomainValidator

class DomainValidatorTest(BaseUnitTest):

  def test_implementation_check(self):
    empty_domain = BaseDomain()
    validator = DomainValidator(empty_domain)
    status, msg = validator.implementation_check()
    self.false(status)
    self.include("generate_inital_state", msg)




from kyoka.value_function.base_action_value_function import BaseActionValueFunction
import numpy as np

class BaseKerasActionValueFunction(BaseActionValueFunction):

  def setUp(self):
    self.model = self.generate_model()
    self.last_metric = 0

  def calculate_value(self, state, action):
    model_input = self.transform_state_action_into_input(state, action)
    return self.predict_value(self.model, model_input)

  def update_function(self, state, action, new_value):
    model_input = self.transform_state_action_into_input(state, action)
    X, Y = map(self.__wrap_data_for_batch, [model_input, new_value])
    metrics = self.model.train_on_batch(X, Y)
    metric = self.fetch_training_metric(metrics)
    delta = metric - self.last_metric
    self.last_metric = metric
    return delta

  def save_model_weights(self, file_path):
    self.model.save_weights(file_path)

  def load_model_weights(self, file_path):
    self.model.load_weights(file_path)

  def provide_data_to_store(self):
    pass

  def receive_data_to_restore(self, restored_data):
    pass

  def generate_model(self):
    err_msg = self.__build_err_msg("generate_model")
    raise NotImplementedError(err_msg)

  def transform_state_action_into_input(self, state, action):
    err_msg = self.__build_err_msg("transform_state_action_into_input")
    raise NotImplementedError(err_msg)

  def predict_value(self, model, X):
    err_msg = self.__build_err_msg("predict_value")
    raise NotImplementedError(err_msg)

  def fetch_training_metric(self, metrics):
    err_msg = self.__build_err_msg("fetch_training_metric")
    raise NotImplementedError(err_msg)

  def __wrap_data_for_batch(self, data):
    return np.array([data])

  def __build_err_msg(self, msg):
    base_msg = "[ {0} ] class does not implement [ {1} ] method"
    return base_msg.format(self.__class__.__name__, msg)




from kyoka.finish_rule.base_finish_rule import BaseFinishRule

class WatchUpdateDelta(BaseFinishRule):

  def __init__(self, patience, minimum_required_delta, log_interval=100):
    BaseFinishRule.__init__(self, log_interval)
    self.patience = patience
    self.minimum_required_delta = minimum_required_delta
    self.no_update_counter = 0
    self.max_delta_memo = []

  def check_condition(self, _iteration_count, deltas):
    max_delta = max([abs(delta) for delta in deltas])
    self.max_delta_memo.append(max_delta)
    if max_delta >= self.minimum_required_delta:
      self.no_update_counter = 0
      self.max_delta_memo = []
    else:
      self.no_update_counter += 1
    return self.no_update_counter >= self.patience

  def generate_progress_message(self, iteration_count, deltas):
    base_msg = "Current iteration count = %d, finish if no update within %d iteration."
    return base_msg % (iteration_count, self.patience - self.no_update_counter)

  def generate_finish_message(self, iteration_count, deltas):
    base_msg = "Update of deltas are less than %f while %d iteration (maimum delta was %f). So stop GPI process"
    maximum_delta = max(self.max_delta_memo)
    return base_msg % (self.minimum_required_delta, self.patience, maximum_delta)




from tests.base_unittest import BaseUnitTest
from sample.ticktacktoe.ticktacktoe_table_value_function import TickTackToeTableValueFunction

class TickTackToeTableValueFunctionTest(BaseUnitTest):

  def setUp(self):
    self.func = TickTackToeTableValueFunction()
    self.func.setUp()

  def test_generate_initial_table(self):
    table = self.func.generate_initial_table()
    self.eq(2**9, len(table))
    self.eq(2**9, len(table[0]))
    self.eq(9, len(table[0][0]))
    self.eq(0, table[0][0][0])

  def test_calculate_value(self):
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("000000100")
    second_player_board = bin2i("000000010")
    state = (first_player_board, second_player_board)
    action = 1
    self.func.table[4][2][0] = 1  # 4 = first_board, 2 = second_board, 0 = log(action, base=2)
    self.eq(1, self.func.calculate_value(state, action))

  def test_update_function(self):
    bin2i = lambda b: int(b, 2)
    first_player_board = bin2i("000000100")
    second_player_board = bin2i("000000010")
    state = (first_player_board, second_player_board)
    action = 1

    self.eq(0, self.func.calculate_value(state, action))

    delta = self.func.update_function(state, action, 1)
    self.eq(1, self.func.calculate_value(state, action))
    self.eq(1, delta)

    delta = self.func.update_function(state, action, 0)
    self.eq(0, self.func.calculate_value(state, action))
    self.eq(-1, delta)




import os

from tests.base_unittest import BaseUnitTest
from kyoka.finish_rule.manual_interruption import ManualInterruption

from nose.tools import raises

class ManualInterruptionTest(BaseUnitTest):

  def setUp(self):
    file_path = self.__generate_tmp_file_path()
    self.rule = ManualInterruption(monitor_file_path=file_path)

  def tearDown(self):
    file_path = self.__generate_tmp_file_path()
    if os.path.isfile(file_path):
      os.remove(file_path)

  def test_satisfy_condition(self):
    file_path = self.__generate_tmp_file_path()
    self.false(self.rule.satisfy_condition(1, "dummy"))
    self.__write_word(file_path, "hoge")
    self.false(self.rule.satisfy_condition(1, "dummy"))
    self.__write_word(file_path, "stop")
    self.true(self.rule.satisfy_condition(1, "dummy"))

  def test_generate_progress_message(self):
    self.eq(None, self.rule.generate_progress_message(5, "dummy"))

  def test_generate_finish_message(self):
    file_path = self.__generate_tmp_file_path()
    msg = self.rule.generate_finish_message(5, "dummy")
    self.include(str(5), msg)
    self.include(file_path, msg)

  def __generate_tmp_file_path(self):
    return os.path.join(os.path.dirname(__file__), "tmp_file_for_manual_interruption_test.tmp")

  def __write_word(self, filepath, word):
    with open(filepath, 'wb') as f:
      f.write(word)

